module CommonMain where

import DictToDictionary
import Print
import List(intersperse,isPrefixOf)
import System(getArgs, getEnv)
import GeneralIO
import General
import IO
import qualified Data.Set as Set
import System(getProgName)
import Dictionary
import Frontend
import Char
import Dict.ErrM
import Print
import Monad
import UTF8
import Command
import Util
import qualified Data.Map as Map

-- import Tokenize
import qualified CTrie

type AnaType = (AT,Bool)

data AT = 
    Normal            |
    NoAnalysis        |
    FilterLexiconNo   |
    FilterLexiconComp
    
gfTypes :: Language a => a -> String
gfTypes l = "types." ++ name l ++ ".gf"

readDicts :: Language a => a -> [FilePath] -> (Bool,Bool,Bool) -> IO (Dictionary,Int)
readDicts l fs (undefcheck,argccheck,unusedcheck) = do output
                                                       readDicts' l fs
 where
  output 
   | length fs > 1 = prErr $ "\nprocessing dictionaries in files: " ++ (unwords fs) 
   | null fs       = prErr $ "\nno dictionary loaded"
   | otherwise     = prErr $ "\nprocessing dictionary in file " ++ (unwords fs)
  readDicts' l [] = return $ (internDict l, 0)
  readDicts' l (f:fs) = do (d,n) <- readDicts' l fs
                           res   <- parseDict l f (undefcheck,argccheck,unusedcheck)
                           case res of
                             Ok (d1,n1) -> return $ (unionDictionary d d1,n+n1)
                             Bad s      -> do prErr s
                                              return (d,n)

prStatistics :: Language a => a -> Int -> IO()
prStatistics l n =
 do let is = size (internDict l)
    prErr $ print_lang l ++ print_paradigms l ++ print_size (n+is,n,is) ++ "\n"

print_lang l = "language id: " ++ (name l) ++ "\n"

print_paradigms l = case paradigmCount l of
                     0 -> "no paradigms\n"
                     1 -> "1 paradigm\n"
                     n -> show n ++ " paradigms\n"

print_size (0,n,isz)  = "no/empty dictionary"
print_size (1,_,_)    = "1 entry"
print_size (sz,n,isz) = show nsz ++ "k entries (e: " ++ show n ++ ", i: " ++ show isz ++ ")"
  where 
        nsz = fromInteger $ round (((fromInteger (toInteger sz))::Double) / 1000)

uName :: Language a => a -> String
uName l = case name l of
	   [] -> []
	   (x:xs) -> toUpper x : xs

commonMain :: Language a => a -> IO ()
commonMain l = do
  xx <- getArgs
  res <- retrieve l xx
  case res of
    Left err -> do prErr $ err
    Right (flags, []) | dictionary_needed flags
    	-> do prg <- getProgName
              prErr $ welcome l
              prErr $ "dictionary file required"
              prErr $ "Usage: " ++ prg ++ " [OPTION...] dictionary_file(s)...\n"
    Right (flags, (_:_)) | is_fullform flags -> do prg <- getProgName
                                                   prErr $ "fullform and dictionary files may not be used at the same time."
    Right (flags, files) -> 
     if is_help flags then do s <- help l ; prErr (welcome l) ; prErr s else if is_version flags then hPutStrLnUTF8 True stdout (welcome l) else
      do prErr $ welcome l
         let undefcheck      = is_undef flags      || is_all flags
             unusedcheck     = is_unused flags     || is_all flags
             argccheck       = is_argc flags       || is_all flags
             dupcheck        = is_duplicated flags || is_all flags
             (compInf,cmode) = (is_compound flags, get_compound flags)
             tokS            = utf8_or_not flags . (get_tokenizer (tokenizer l) flags) . decodeUTF8
         when compInf $ prErr (case (composition l) of 
                                Nothing -> "\nCompounding activated: the naive method. mode: " ++ (pr_comp cmode)
                                _       -> "\nCompounding activated: the default method. mode: " ++ (pr_comp cmode))
         (d,n) <- if (dictionary_needed flags) then 
                      do (dict,n1) <- readDicts l files (undefcheck,argccheck,unusedcheck)
                         return (apply_encoding l flags dict,n1)
                   else 
                       return (emptyDict,0)
         case flags of
           fs | invalid_tokenizer fs -> prErr $ "unknown tokenizer: " ++ (get_tokenizer_name fs)
           fs | invalid_quality   fs -> prErr $ "unknown quality argument: " ++ (get_quality fs)
           fs | is_quality fs -> do when dupcheck   $ check_lemma_duplication d
                                    when (is_test fs || is_all fs) $ tester (testBench l) (testdata d) 
                                    when (is_dict fs || is_all fs) $ print_duplicates d (dup_id_exceptions l)
                                    return ()
           fs | elem Infl     fs ->  do prErr $ "[ FM inflection mode ]"
                                        s <- imode l
                                        output_write fs s
           fs | elem Synth   fs  -> do prErr $ "[ FM synthesiser mode ]\n"
			               CTrie.buildTrieDictSynt d False
                                       prStatistics l n
                                       s <- synthesiser l
                                       output_write fs s
           fs | elem Tag fs -> do --CTrie.buildTrieDict (isComp l) d False dupcheck
                                  build_trie l fs d
                                  prStatistics l n
			          s <- posify l tokS (analysis (compInf,cmode) (composition l))
                                  output_write fs s
           fs | is_mode fs  -> do --CTrie.buildTrieDict (isComp l) d False dupcheck
                                  build_trie l fs d
                                  prStatistics l n
			          run l (is_fullform fs) tokS (analysis (compInf,cmode) (composition l)) (get_mode fs)
           fs | is_printer fs -> 
                output_write fs $
                  case printer fs of
                   (Just "core")      -> unlines (paradigmNames l)
                   (Just "paradigms") -> prParadigms l $ apply_encoding l flags (dictionary [f xs | (p,(xs,f)) <- Map.toList (paradigms l)])
                   (Just "tagset")    -> prTagset l    $ apply_encoding l flags (dictionary [f xs | (p,(xs,f)) <- Map.toList (paradigms l)])
                   (Just "extract")   -> prExtract l
                   (Just "newlex")  -> prNewDictionary d
           --        (Just "json")    -> prJSON d
	           (Just "lex")     -> prJSON d --prFullFormLex (dict2fullform d False)
	           (Just "tabbedlex") -> prTabbedLex d
	           (Just "words")   -> prWordlist (dict2fullform d)
	           (Just "tables")  -> prDictionary d
	           (Just "gf")      -> "-- machine-generated GF file\n\n" ++
		                        "include " ++ (gfTypes l) ++ " ;\n\n" ++ 
		                         (prGF d)
	 --          (Just "gfr")     ->  "-- machine-generated GF file\n\n" ++
	 --	                        "include " ++ (gfTypes l) ++ " ;\n\n" ++ 
	--	                        prGFRes d
    	           (Just "latex")   -> prLatex d
	           (Just "xml")     -> prXML d
                   (Just "sfst")    -> prSFST d
                   (Just "sfstheader") -> prSFSTHEAD d
                   (Just "sfstlex")  -> prSFSTLEX d
	           (Just "lexc")     -> prLEXC d
	           (Just "xfst")     -> prXFST d
                   (Just "hundict")  -> prHunDict d
                   (Just "hunaffix") -> prHunAffix d
	           (Just "sql")      -> prSQL d
                   (Just "lmf")      -> prLMF (name l) d
                   (Just  x)        -> error $ "unknown printer: " ++ x
                   Nothing          -> error $ "Internal error. This is a bug."
           fs -> do build_trie l fs d
                    prStatistics l n
		    run l (is_fullform fs) tokS (analysis (compInf,cmode) (composition l)) []

build_trie :: Language l => l -> [Flag] -> Dictionary -> IO ()
build_trie l fs d = CTrie.buildTrieDict (isComp l) d False
-- | is_fullform fs = CTrie.buildTrie (get_fullform fs) (isComp l)
-- | otherwise      = 

data Stats = Stats {
		    totalWords :: Int,
		    coveredWords :: Int
		   }

initStats :: Stats
initStats = Stats { totalWords = 0, coveredWords = 0 }

utf8_or_not :: [Flag] -> [Tok] -> [Tok]
utf8_or_not fs
 | is_fullform fs = map encodeUTF8Tok
 | otherwise      = id

posify :: Language a => a -> (String -> [Tok]) -> (String -> [[String]]) -> IO String
posify l lexer f = do 
  s' <- hGetContents stdin
  let ts  = lexer s'
      tss = get_sentences ts
  return $  concat [printResult (map anapos x) | x <- tss]
 where 
  printResult [] = []
  printResult xs = "{" ++ (unwords xs) ++ "}\n"
  anapos t = 
   case t of
    (P s)  -> "(\"" ++ esc s ++ "\",spec)"
    (PD s) -> "(\"" ++ esc s ++ "\",num/spec)"
    (D s) -> "(\"" ++ esc s ++ "\",num)"
    (W s) ->  case f s of
                [] -> "(\"" ++ esc s ++ "\",)"
                xs -> "(\"" ++ esc s ++ "\"," ++ prResult xs ++ ")"
    (A (u,l)) -> case (f u) ++ (f l) of
                  [] -> "(\"" ++ esc u ++ "\",)"
                  xs -> "(\"" ++ esc u ++ "\"," ++ prResult xs ++ ")"
    (AA (u,m,l)) -> case (f u) ++ (f m) ++ (f l) of
                     [] -> "(\"" ++ esc u ++ "\",)"
                     xs -> "(\"" ++ esc u ++ "\"," ++ prResult xs ++ ")"
  prResult :: [[String]] -> String
  prResult xs  = concat $ intersperse "|" $ filter (not.null) (map filter_analysis xs)
  esc [] = []
  esc ('\"':xs) = '\\':'\"':esc xs
  esc (x:xs) = x:esc xs
  filter_analysis :: [String] -> String
  filter_analysis [s] = get_pos s ++ " " ++ get_param s
  filter_analysis _   = []
  get_pos s@(x:xs)
   | isPrefixOf "pos\":" s = case span (/= '\"') (drop 6 s) of
                               (r,_) -> r
   | otherwise             = get_pos xs
  get_param s@(x:xs)
   | isPrefixOf "param\":" s = case span (/= '\"') (drop 8 s) of
                               (r,_) -> r
   | otherwise             = get_param xs

get_sentences :: [Tok] -> [[Tok]]
get_sentences xs = gets xs []
   where gets [] s = [reverse s]
         gets (c:cs) s
          | isMajor c = (reverse (c:s)):gets cs []
          | otherwise = gets cs (c:s)

isMajor (P [c]) = elem c ".?!"
isMajor _     = False

run :: Language a => a -> Bool -> (String -> [Tok]) -> (String -> [[String]]) -> String -> IO ()
run l b t f "fail"      = run' l t f (NoAnalysis,b)        initStats >> return ()
run l b t f "lexfail"   = run' l t f (FilterLexiconNo,b)   initStats >> return ()
run l b t f "lexcomp"   = run' l t f (FilterLexiconComp,b) initStats >> return ()
run l b t f _           = do
                          st <- run' l t f (Normal,b) initStats
		          prErr $ "Total words:   " ++ show (totalWords st)
		          prErr $ "Covered words: " ++ show (coveredWords st)

run' :: Language a => a -> (String -> [Tok]) -> (String -> [[String]]) -> AnaType -> Stats -> IO Stats
run' l tokenizer f t@(a,_) st = 
 do 
   b <- hIsEOF stdin
   if b then return st 
    else do 
          s <- hGetLine stdin
	  case a of
	    FilterLexiconNo   -> 
             analyze l t f (tokenizer s) st >>= run' l tokenizer f t
	    FilterLexiconComp -> 
	     analyze l t f (tokenizer s) st >>= run' l tokenizer f t
	    _                 ->  
	     analyze l t f (tokenizer s) st >>= run' l tokenizer f t

word_tokens :: [Tok] -> [String]
word_tokens xs = [s | (W s) <- xs]

hPutStrLnUTF8 b h s = hPutStrLn h $ (if b then id else encodeUTF8) s

hPutStrUTF8 b  h s = hPutStr h   $ (if b then id else encodeUTF8) s

analyze :: Language a => a -> AnaType -> (String -> [[String]]) -> [Tok] -> Stats -> IO Stats
analyze _ (FilterLexiconNo,b) f ((W x):xs) st
     | or $ map (\ys -> length ys > 0) (concat (map f (word_tokens xs))) = 
	 do
	 hPutStrLnUTF8 b stdout $ unwords (x:(word_tokens xs))
	 return st
     | otherwise = return st
analyze _ (FilterLexiconComp,b) f ((W x):xs) st
     | or $ map (\ys -> length ys > 1) (concat (map f (word_tokens xs))) = 
	 do
	 hPutStrLnUTF8 b stdout $ unwords (x:(word_tokens xs))
	 return st
     | otherwise = return st
analyze _ _ _  [] st = return st
analyze l t f (s:ss) st =
   case s of
    (P s) ->   
        do case t of 
	    (Normal,b)    
             -> do hPutStrLnUTF8 b stdout $ "{\"" ++ s ++ "\":\"-Symb-\"}\n"
		   analyze l t f ss st
            _ -> analyze l t f ss st
    (PD s) ->   
        do case t of 
	    (Normal,b)    
             -> do hPutStrLnUTF8 b stdout $ "{\"" ++ s ++ "\":\"-Num/Symb-\"}\n"
		   analyze l t f ss st
            _ -> analyze l t f ss st
    (D s) ->
        do case t of 
            (Normal,b) 
             -> do hPutStrLnUTF8 b stdout $ "{\"" ++ s ++ "\":\"-Num-\"}\n"
		   analyze l t f ss st
	    _ -> analyze l t f ss st
    (W s) ->  
	case f s of
        [] -> do prAnalysis l t s []
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st 
			            })
        xs -> do prAnalysis l t s xs
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st + 1 
			            })
    (A (s,ls)) ->  
	case (f s) ++(f ls) of
        [] -> do prAnalysis l t s []
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st 
			            })
        xs -> do prAnalysis l t s xs
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st + 1 
			            })
    (AA (s,m,ls)) ->  
	case (f s) ++ (f m) ++(f ls) of
        [] -> do prAnalysis l t s []
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st 
			            })
        xs -> do prAnalysis l t s xs
	         analyze l t f ss (st {
			             totalWords = totalWords st + 1,
			                          coveredWords = coveredWords st + 1 
			            })


prAnalysis :: Language a => a -> AnaType -> String -> [[String]] -> IO()
prAnalysis l (Normal,b)   s   [] = case wordGuesser l s of
                                 (x:xs) ->  do mapM_ (hPutStrUTF8 b stdout) ["{\"", s, "\":[\"-Guess-\",\n"]
                                               hPutStrLnUTF8 b stdout $ prA (map (:[]) (x:xs))
                                               hPutStrLnUTF8 b stdout "}\n"
                                 []     ->
                                   hPutStrLnUTF8 b stdout $ concat ["{\"", s, "\":\"-Unknown-\"}\n"]
prAnalysis _ (Normal,b)   s   xs = do mapM_ (hPutStrUTF8 b stdout) ["{\"", s, "\":[\n"]
                                      hPutStrLnUTF8 b stdout $ prA xs
                                      hPutStrLnUTF8 b stdout "]}\n"
prAnalysis _ (NoAnalysis,b) s []    = hPutStrLnUTF8 b stdout $ s
prAnalysis _ (NoAnalysis,b) s _     = return ()
prAnalysis _ _      _ _             = return ()

prA xs =  concat $ intersperse ",\n" [s | s <- map pr xs]

pr []  = []
pr [x] = x
pr xs  = "{\"composite\":" ++ "[" ++ (concat (intersperse "," xs)) ++ "]"

welcome :: Language a => a -> String
welcome l = "FM 2.1 (c) M. Forsberg & A. Ranta, 2008, under GNU GPL"
 
